option add *Menu*selectColor 		forestgreen
option add *Radiobutton*selectColor 	forestgreen
option add *Checkbutton*selectColor 	forestgreen
option add *Entry.background 		gray70
option add *Entry.relief       		sunken 
option add *borderWidth        		1
option add *highlightThickness 		0
option add *font               		{Helvetica 8}
option add *Menu*tearOff       		0

canvas    .c  -xscrollcommand {.sb set} -yscrollcommand {.sr set} 
frame     .c.f 
scrollbar .sr -orient vertical   -command {.c yview}
scrollbar .sb -orient horizontal -command {.c xview}

pack .sb -side bottom -fill x    -expand 0 -anchor s
pack .sr -side right  -fill y    -expand 0 -anchor e
pack .c  -side left   -fill both -expand 1 -anchor n

.c create window 0 0 -anchor nw -window .c.f

proc mtrace_callback {fd src dst} {
	if [winfo exists .mtrace-$src-$dst] {
		.mtrace-$src-$dst.t insert end [read $fd]
	} else {
		# The user has closed the mtrace window before the trace has completed
		close $fd
	}
	if [eof $fd] {
		close $fd
	}
}

proc mtrace {src dst} {
	global CNAME group_addr
	regsub {.*@([0-9]+.[0-9]+.[0-9]+.[0-9]+)} $CNAME($src) {\1} src_addr
	regsub {.*@([0-9]+.[0-9]+.[0-9]+.[0-9]+)} $CNAME($dst) {\1} dst_addr

	toplevel  .mtrace-$src-$dst
	text      .mtrace-$src-$dst.t -background white -font "Courier 8" -wrap none \
	                              -yscrollcommand ".mtrace-$src-$dst.sr set" \
	                              -xscrollcommand ".mtrace-$src-$dst.sb set"
	scrollbar .mtrace-$src-$dst.sb -command ".mtrace-$src-$dst.t xview" -orient horizontal
	scrollbar .mtrace-$src-$dst.sr -command ".mtrace-$src-$dst.t yview" -orient vertical
	pack .mtrace-$src-$dst.sb -fill x    -expand 0 -side bottom -anchor s
	pack .mtrace-$src-$dst.sr -fill y    -expand 0 -side right  -anchor e 
	pack .mtrace-$src-$dst.t  -fill both -expand 1 -side left   -anchor w

	wm title .mtrace-$src-$dst "mtrace from $src_addr to $dst_addr via group $group_addr"

	set fd [open "|mtrace $src_addr $dst_addr $group_addr" "r"]
	fconfigure $fd -blocking 0
	fileevent  $fd readable "mtrace_callback $fd $src $dst"
}

toplevel .chart_popup       -bg black
label    .chart_popup.text  -bg lavender -justify left
pack .chart_popup.text -side top -anchor w -fill x
wm transient        .chart_popup .
wm withdraw         .chart_popup
wm overrideredirect .chart_popup true

proc chart_popup_show {window} {
	global chart_popup_src chart_popup_dst chart_popup_id NAME CNAME

	set name_from $NAME($chart_popup_src($window))
	if {[string compare $name_from ""] == 0} {
		set name_from $CNAME($chart_popup_src($window))
	}
	if {[string compare $name_from ""] == 0} {
		set name_from $chart_popup_src($window)
	}

	set name_to $NAME($chart_popup_dst($window))
	if {[string compare $name_to ""] == 0} {
		set name_to $CNAME($chart_popup_dst($window))
	}
	if {[string compare $name_to ""] == 0} {
		set name_to $chart_popup_dst($window)
	}

	.chart_popup.text  configure -text "From: $name_from\nTo: $name_to"
	# Beware! Don't put the popup under the cursor! Else we get window enter
	# for .help and leave for $window, making us hide_help which removes the
	# help window, giving us a window enter for $window making us popup the
	# help again.....
	if {[winfo width $window] > [winfo height $window]} {
	    set xpos [expr [winfo pointerx $window] + 10]
	    set ypos [expr [winfo rooty    $window] + [winfo height $window] + 4]
	} else {
	    set xpos [expr [winfo rootx    $window] + [winfo width $window] + 4]
	    set ypos [expr [winfo pointery $window] + 10]
	}
	
	wm geometry  .chart_popup +$xpos+$ypos
	set chart_popup_id [after 100 wm deiconify .chart_popup]
	raise .chart_popup $window
}

proc chart_popup_hide {window} {
	global chart_popup_id 
	if {[info exists chart_popup_id]} { 
		after cancel $chart_popup_id
	}
	wm withdraw .chart_popup
}

proc chart_popup_add {window src dst} {
	global chart_popup_src chart_popup_dst 
	set chart_popup_src($window) $src
	set chart_popup_dst($window) $dst
	bind $window <Enter>    "+chart_popup_show $window"
	bind $window <Leave>    "+chart_popup_hide $window"
}

proc chart_add {ssrc} {
	frame  .c.f.$ssrc
	frame  .c.f.$ssrc.f
	button .c.f.$ssrc.l -width 25 -font "Helvetica 10" -text $ssrc -padx 0 -pady 0 -anchor w -relief flat -command "toggle_stats $ssrc"
	pack   .c.f.$ssrc.f -expand 0 -anchor e -side right
	pack   .c.f.$ssrc.l -expand 1 -anchor w -fill x -side left
	pack   .c.f.$ssrc   -expand 1 -anchor n -fill x -side top
	foreach s [pack slaves .c.f] {
		regsub {.c.f.(.*)} $s {\1} s
		button .c.f.$s.f.$ssrc -width 4 -font "Helvetica 10" -text "" -background white -padx 0 -pady 0 -command "mtrace $s $ssrc"
		pack   .c.f.$s.f.$ssrc -expand 0 -side left
		chart_popup_add .c.f.$s.f.$ssrc $s $ssrc
		if {![winfo exists .c.f.$ssrc.f.$s]} {
			button .c.f.$ssrc.f.$s -width 4 -font "Helvetica 10" -text "" -background white -padx 0 -pady 0 -command "mtrace $ssrc $s"
			pack   .c.f.$ssrc.f.$s -expand 0 -side left
			chart_popup_add .c.f.$ssrc.f.$s $ssrc $s
		}
	}
	update
	.c configure -scrollregion "0.0 0.0 [winfo width .c.f] [winfo height .c.f]"
}

proc chart_remove {ssrc} {
	destroy .c.f.$ssrc
	foreach s [pack slaves .c.f] {
		regsub {.c.f.(.*)} $s {\1} s
		destroy .c.f.$s.f.$ssrc
	}
	update
	.c configure -scrollregion "0.0 0.0 [winfo width .c.f] [winfo height .c.f]"
}

proc chart_label {ssrc label} {
	.c.f.$ssrc.l configure -text $label
}

proc chart_set {src dst val} {
	set txtval "$val%"
	if     {$val ==  0} {
		set colour "#00ff00"
	} elseif {$val ==  1} {
		set colour "#2fff00"
	} elseif {$val ==  2} {
		set colour "#4fff00"
	} elseif {$val ==  3} {
		set colour "#6fff00"
	} elseif {$val ==  4} {
		set colour "#afff00"
	} elseif {$val ==  5} {
		set colour "#bfff00"
	} elseif {$val ==  6} {
		set colour "#dfef00"
	} elseif {$val ==  7} {
		set colour "#ffdf00"
	} elseif {$val ==  8} {
		set colour "#ffc500"
	} elseif {$val ==  9} {
		set colour "#ffb500"
	} elseif {$val == 10} {
		set colour "#ffa500"
	} elseif {$val == 11} {
		set colour "#ff9000"
	} elseif {$val == 12} {
		set colour "#ff8000"
	} elseif {$val == 13} {
		set colour "#ff7000"
	} elseif {$val == 14} {
		set colour "#ff6000"
	} elseif {$val == 15} {
		set colour "#ff5000"
	} elseif {$val == 16} {
		set colour "#ff4000"
	} elseif {$val == 17} {
		set colour "#ff3000"
	} elseif {$val == 18} {
		set colour "#ff2000"
	} elseif {$val == 19} {
		set colour "#ff1000"
	} elseif {$val < 100} {
		set colour "#ff0000"
	} elseif {$val == 101} {
		# No information
		set colour white
		set txtval " "
	} elseif {$val == 102} {
		# Empty RR
		set colour lavender
		set txtval "-"
	} else {
		# Should never happen
		set colour black
		set txtval " "
	}
	.c.f.$src.f.$dst configure -background $colour -text "$txtval"
}

proc rr_empty {ssrc} {
	# An empty RR was received from $ssrc, indicating that this source
	# cannot hear anyone else. 
	global SSRC
	foreach s [array names SSRC] {
		chart_set $s $ssrc 102
	}
}

proc ssrc_delete {ssrc} {
	global CNAME NAME EMAIL PHONE LOC TOOL SSRC CODEC PCKTS_RECV PCKTS_LOST LOSS_FROM_ME LOSS_TO_ME JITTER 
	unset         SSRC($ssrc)
	unset        CNAME($ssrc)
	unset         NAME($ssrc)
	unset        EMAIL($ssrc)
	unset        PHONE($ssrc)
	unset          LOC($ssrc)
	unset         TOOL($ssrc)
	unset        CODEC($ssrc)
	unset   PCKTS_RECV($ssrc)
	unset   PCKTS_LOST($ssrc)
	unset LOSS_FROM_ME($ssrc)
	unset   LOSS_TO_ME($ssrc)
	unset       JITTER($ssrc)
	chart_remove $ssrc
}

proc ssrc_exists {ssrc} {
	global CNAME NAME EMAIL PHONE LOC TOOL SSRC CODEC PCKTS_RECV PCKTS_LOST LOSS_FROM_ME LOSS_TO_ME JITTER 
	set         SSRC($ssrc) $ssrc
	set        CNAME($ssrc) ""
	set         NAME($ssrc) ""
	set        EMAIL($ssrc) ""
	set        PHONE($ssrc) ""
	set          LOC($ssrc) ""
	set         TOOL($ssrc) ""
	set        CODEC($ssrc) ""
	set   PCKTS_RECV($ssrc) 0
	set   PCKTS_LOST($ssrc) 0
	set LOSS_FROM_ME($ssrc) 101
	set   LOSS_TO_ME($ssrc) 101
	set       JITTER($ssrc) 0
	chart_add $ssrc
}

proc set_cname {ssrc cname} {
	global CNAME NAME
	if {[string compare $CNAME($ssrc) ""] == 0} {
		chart_label $ssrc $cname
	}
	set CNAME($ssrc) $cname
}

proc set_name {ssrc name} {
	global NAME
	set NAME($ssrc) $name
	chart_label $ssrc $name
}

proc set_email {ssrc email} {
	global EMAIL
	set EMAIL($ssrc) $email
}

proc set_phone {ssrc phone} {
	global PHONE
	set PHONE($ssrc) $phone
}

proc set_loc {ssrc loc} {
	global LOC
	set LOC($ssrc) $loc
}

proc set_tool {ssrc tool} {
	global TOOL
	set TOOL($ssrc) $tool
}

proc window_stats {ssrc} {
	regsub -all {[\. ]} $ssrc {-} foo
	return .stats$foo
}

proc stats_add_field {widget label watchVar} {
    frame $widget -relief sunk 
    label $widget.l -text $label  -anchor w
    label $widget.w -textvariable $watchVar 
    pack $widget -side top -fill x -expand 1 
    pack $widget.l -side left  -fill x -expand 1
    pack $widget.w -side right 
}

proc averageCharacterWidth {font} {
    set sample "1234567890ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
    set slen [string length $sample]
    set wpc  [expr [font measure $font $sample] / $slen + 1]
    return $wpc
}

# constrain_window - window, font, characters wide, characters high
proc constrain_window {w font cW cH} {
	set wpc [averageCharacterWidth $font]
	set hpc [font metrics $font -ascent]
       	# Calculate dimensions
	set width [expr $cW * $wpc]
	set height [expr $cH * $hpc]
	wm geometry $w [format "%sx%s" $width $height]
	set dummy ""
} 

proc toggle_stats {ssrc} {
    set win [window_stats $ssrc]
    if {[winfo exists $win]} {
	destroy $win
    } else {
	global stats_pane
	# Window does not exist so create it
	toplevel $win 

	frame $win.df 
	pack  $win.df -fill x

	global NAME EMAIL PHONE LOC NOTE CNAME TOOL SSRC CODEC PCKTS_RECV PCKTS_LOST LOSS_FROM_ME LOSS_TO_ME JITTER 
	stats_add_field $win.df.1  "Name: "     		NAME($ssrc)
	stats_add_field $win.df.2  "Email: "    		EMAIL($ssrc)
	stats_add_field $win.df.3  "Phone: "    		PHONE($ssrc)
	stats_add_field $win.df.4  "Location: " 		LOC($ssrc)
	stats_add_field $win.df.5  "Note: "     		NOTE($ssrc)
	stats_add_field $win.df.6  "Tool: "     		TOOL($ssrc)
	stats_add_field $win.df.7  "CNAME: "    		CNAME($ssrc)
	stats_add_field $win.df.8  "SSRC: "     		SSRC($ssrc)
	stats_add_field $win.df.9  "Payload type: "           	CODEC($ssrc)

# Window Magic
	frame  $win.dis 
	button $win.dis.b  -text "Dismiss" -command "destroy $win"
	pack   $win.dis   -side bottom -anchor s -fill x -expand 1
	pack   $win.dis.b -side right -anchor e -padx 2 -pady 2
	wm title $win "Participant $NAME($ssrc)"
	wm resizable $win 1 0
	constrain_window $win {Helvetica 8} 36 30
    }
}

